from __future__ import print_function
from importlib.util import LazyLoader
from re import L
import torch
import torch.nn as nn
import torch.optim as optim
from zmq import device
from sparselearning.snip import SNIP, GraSP
import numpy as np
import math

def add_sparse_args(parser):
    parser.add_argument('--sparse', action='store_true', help='Enable sparse mode. Default: True.')
    parser.add_argument('--fix', action='store_true', help='Fix sparse connectivity during training. Default: True.')
    parser.add_argument('--sparse_init', type=str, default='ERK', help='sparse initialization')
    parser.add_argument('--growth', type=str, default='random', help='Growth mode. Choose from: momentum, random, random_unfired, and gradient.')
    parser.add_argument('--death', type=str, default='magnitude', help='Death mode / pruning mode. Choose from: magnitude, SET, threshold.')
    parser.add_argument('--redistribution', type=str, default='none', help='Redistribution mode. Choose from: momentum, magnitude, nonzeros, or none.')
    parser.add_argument('--death-rate', type=float, default=0.50, help='The pruning rate / death rate.')
    parser.add_argument('--density', type=float, default=0.05, help='The density of the overall sparse network.')
    parser.add_argument('--update_frequency', type=int, default=100, metavar='N', help='how many iterations to train between parameter exploration')
    parser.add_argument('--decay-schedule', type=str, default='cosine', help='The decay schedule for the pruning rate. Default: cosine. Choose from: cosine, linear.')
    parser.add_argument('--flow_preservation', type=str, default='True', help='Does flow preservation after ER init')



def get_pyramid_sparsity(model, density):
    num_params = 0
    layerwise_params = []
    for name, m in model.named_modules():
        if isinstance(m, nn.Linear) or isinstance(m, nn.Conv2d):
            num_params += m.weight.numel()
            layerwise_params.append(m.weight.numel())
    layerwise_params = layerwise_params[::-1]
    layerwise_params.append(-num_params*density)
    roots = np.roots(np.array(layerwise_params))
    for r in roots:
        if r < 1 and r > 0 and np.imag(r) == 0:
            layerwise_sparsities = np.arange(1, len(layerwise_params) + 1)
            return np.real(r) ** layerwise_sparsities


class CosineDecay(object):
    def __init__(self, death_rate, T_max, eta_min=0.005, last_epoch=-1):
        self.sgd = optim.SGD(torch.nn.ParameterList([torch.nn.Parameter(torch.zeros(1))]), lr=death_rate)
        self.cosine_stepper = torch.optim.lr_scheduler.CosineAnnealingLR(self.sgd, T_max, eta_min, last_epoch)

    def step(self):
        self.cosine_stepper.step()

    def get_dr(self):
        return self.sgd.param_groups[0]['lr']

class LinearDecay(object):
    def __init__(self, death_rate, factor=0.99, frequency=600):
        self.factor = factor
        self.steps = 0
        self.frequency = frequency

    def step(self):
        self.steps += 1

    def get_dr(self, death_rate):
        if self.steps > 0 and self.steps % self.frequency == 0:
            return death_rate*self.factor
        else:
            return death_rate


class Masking(object):
    def __init__(self, optimizer, death_rate=0.3, growth_death_ratio=1.0, death_rate_decay=None, death_mode='magnitude', growth_mode='momentum', redistribution_mode='momentum', threshold=0.001, args=False, train_loader=False):
        growth_modes = ['random', 'momentum', 'momentum_neuron', 'gradient']
        if growth_mode not in growth_modes:
            print('Growth mode: {0} not supported!'.format(growth_mode))
            print('Supported modes are:', str(growth_modes))

        self.train_loader = train_loader
        self.args = args
        self.device = torch.device("cuda")
        self.growth_mode = growth_mode
        self.death_mode = death_mode
        self.growth_death_ratio = growth_death_ratio
        self.redistribution_mode = redistribution_mode
        self.death_rate_decay = death_rate_decay
        self.flow_preservation_flag = self.args.flow_preservation
        self.model_type = args.model
        self.masks = {}
        self.modules = []
        self.names = []
        self.optimizer = optimizer

        # stats
        self.name2zeros = {}
        self.num_remove = {}
        self.name2nonzeros = {}
        self.death_rate = death_rate
        self.steps = 0

        # if fix, then we do not explore the sparse connectivity
        if self.args.fix: self.prune_every_k_steps = None
        else: self.prune_every_k_steps = self.args.update_frequency

    def init(self, mode='ERK', density=0.05, erk_power_scale=1.0):
        self.density = density
        if mode == 'global_magnitude':
            print('initialize by global magnitude')
            weight_abs = []
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    weight_abs.append(torch.abs(weight))

            # Gather all scores in a single vector and normalise
            all_scores = torch.cat([torch.flatten(x) for x in weight_abs])
            num_params_to_keep = int(len(all_scores) * self.density)

            threshold, _ = torch.topk(all_scores, num_params_to_keep, sorted=True)
            acceptable_score = threshold[-1]

            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    self.masks[name] = ((torch.abs(weight)) >= acceptable_score).float()

        elif mode == 'ER_init':
            print("DEBUGGING... ENTER ER INIT MASKING")
            self.baseline_nonzero = 0
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    if len(weight.shape) == 2 or len(weight.shape) == 4:
                        curr_mask = torch.zeros_like(weight)
                        curr_mask = curr_mask.bernoulli_(p = density)
                        self.masks[name][:] = curr_mask
                        self.baseline_nonzero += (self.masks[name] != 0).sum().int().item()

            self.apply_mask()

        elif mode == 'snip':
            print('initialize by snip')
            layer_wise_sparsities = SNIP(self.module, self.density, self.train_loader, self.device)
            # re-sample mask positions
            for sparsity_, name in zip(layer_wise_sparsities, self.masks):
                self.masks[name][:] = torch.empty(self.masks[name].shape).bernoulli_(p=1 - sparsity_).float().data.cuda()

        elif mode == 'GraSP':
            print('initialize by GraSP')
            layer_wise_sparsities = GraSP(self.module, self.density, self.train_loader, self.device)
            # re-sample mask positions
            for sparsity_, name in zip(layer_wise_sparsities, self.masks):
                self.masks[name][:] = torch.empty(self.masks[name].shape).bernoulli_(p=1 - sparsity_).float().data.cuda()

        elif mode == 'pyramid_sparsity':
            print('initalize by pyramid sparsity')
            layer_wise_sparsities = get_pyramid_sparsity(self.module, self.density)
            for density_, name in zip(layer_wise_sparsities, self.masks):
                print(density_)
                self.masks[name][:] = torch.empty(self.masks[name].shape).bernoulli_(p = density_).float().data.cuda()

        elif mode == 'uniform_plus':
            print('initialize by uniform+')
            total_params = 0
            for name, weight in self.masks.items():
                total_params += weight.numel()
            total_sparse_params = total_params * self.density

            # remove the first layer
            total_sparse_params = total_sparse_params - self.masks['conv.weight'].numel()
            self.masks.pop('conv.weight')

            if self.density < 0.2:
                total_sparse_params = total_sparse_params - self.masks['fc.weight'].numel() * 0.2
                self.density = float(total_sparse_params / total_params)

                for module in self.modules:
                    for name, weight in module.named_parameters():
                        if name not in self.masks: continue
                        if name != 'fc.weight':
                            self.masks[name][:] = (torch.rand(weight.shape) < self.density).float().data.cuda()
                        else:
                            self.masks[name][:] = (torch.rand(weight.shape) < 0.2).float().data.cuda()
            else:
                for module in self.modules:
                    for name, weight in module.named_parameters():
                        if name not in self.masks: continue
                        self.masks[name][:] = (torch.rand(weight.shape) < self.density).float().data.cuda()

        elif mode == 'uniform':
            print('initialize by uniform')
            self.baseline_nonzero = 0
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    self.masks[name][:] = (torch.rand(weight.shape) < self.density).float().data.cuda()
                    self.baseline_nonzero += weight.numel()*density

        
        elif mode == 'ER_init_synflow':
            print("DEBUGGING... ENTER ER INIT GLOBAL MASKING SYNFLOW")

            p = 'synflow'
            model_name = 'vgg' if self.model_type == 'vgg16' else 'resnet'
            target_sparsity = density
            device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        
            mask_list = torch.load("./pretrained-models/cifar-{}-{}-{}/model_mask_list_prunable_params.pt".format(model_name, p, target_sparsity), map_location=device)
            sparsity_list = []
            for n, m in mask_list:
                if len(m.shape) == 2 or len(m.shape) == 4:
                    curr_s = m.sum() / m.numel()
                    sparsity_list.append(curr_s)

            layer = 0
            self.baseline_nonzero = 0
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    if len(weight.shape) == 2 or len(weight.shape) == 4:
                        s = sparsity_list[layer]
                        curr_mask = torch.zeros_like(weight)
                        curr_mask = curr_mask.bernoulli_(p = s)
                        self.masks[name][:] = curr_mask
                        self.baseline_nonzero += (self.masks[name] != 0).sum().int().item()

                        layer += 1

            self.apply_mask()

        elif mode == 'ER_init_imp':
            print("DEBUGGING... ENTER ER INIT GLOBAL MASKING IMP")

            p = 'imp'
            model_name = 'vgg' if self.model_type == 'vgg16' else 'resnet'
            target_sparsity = density
            device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        
            mask_list = torch.load("./pretrained-models/cifar-{}-{}-{}/model_mask_list_prunable_params.pt".format(model_name, p, target_sparsity), map_location=device)
            sparsity_list = []
            for n, m in mask_list:
                if len(m.shape) == 2 or len(m.shape) == 4:
                    curr_s = m.sum() / m.numel()
                    sparsity_list.append(curr_s)

            layer = 0
            self.baseline_nonzero = 0
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    if len(weight.shape) == 2 or len(weight.shape) == 4:
                        s = sparsity_list[layer]
                        curr_mask = torch.zeros_like(weight)
                        curr_mask = curr_mask.bernoulli_(p = s)
                        self.masks[name][:] = curr_mask
                        self.baseline_nonzero += (self.masks[name] != 0).sum().int().item()
                        
                        layer += 1

            self.apply_mask()


        elif mode == 'ER_same_params_per_layer':
            total_params = 0
            l = 0
            sparsity_list = []
            for name, weight in self.masks.items():
                total_params += weight.numel()
                l += 1
            X = density * total_params / l
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue            
                    if X / self.masks[name].numel() < 1:
                        sparsity_list.append(X / self.masks[name].numel())
                    else: 
                        sparsity_list.append(1)
            print(sparsity_list)
            layer = 0
            self.baseline_nonzero = 0
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    if len(weight.shape) == 2 or len(weight.shape) == 4:
                        s = sparsity_list[layer]
                        print(s)
                        curr_mask = torch.zeros_like(weight)
                        curr_mask = curr_mask.bernoulli_(p = s)
                        self.masks[name][:] = curr_mask
                        self.baseline_nonzero += (self.masks[name] != 0).sum().int().item()
                        
                        layer += 1

            self.apply_mask()
        

        elif mode == 'ER':
            print('initialize by ER')
            total_params = 0
            for name, weight in self.masks.items():
                total_params += weight.numel()
            is_epsilon_valid = False

            dense_layers = set()
            while not is_epsilon_valid:

                divisor = 0
                rhs = 0
                raw_probabilities = {}
                for name, mask in self.masks.items():
                    n_param = np.prod(mask.shape)
                    n_zeros = n_param * (1 - self.density)
                    n_ones = n_param * self.density

                    if name in dense_layers:
                        rhs -= n_zeros
                    else:
                        rhs += n_ones
                        raw_probabilities[name] = (
                                                          np.sum(mask.shape[:2]) / np.prod(mask.shape[:2])
                                                  ) ** erk_power_scale
                        divisor += raw_probabilities[name] * n_param
                epsilon = rhs / divisor
                max_prob = np.max(list(raw_probabilities.values()))
                max_prob_one = max_prob * epsilon
                if max_prob_one > 1:
                    is_epsilon_valid = False
                    for mask_name, mask_raw_prob in raw_probabilities.items():
                        if mask_raw_prob == max_prob:
                            print(f"Sparsity of var:{mask_name} had to be set to 0.")
                            dense_layers.add(mask_name)
                else:
                    is_epsilon_valid = True

            density_dict = {}
            total_nonzero = 0.0
            # With the valid epsilon, we can set sparsities of the remaning layers.
            for name, mask in self.masks.items():
                n_param = np.prod(mask.shape)
                if name in dense_layers:
                    density_dict[name] = 1.0
                else:
                    probability_one = epsilon * raw_probabilities[name]
                    density_dict[name] = probability_one
                print(
                    f"layer: {name}, shape: {mask.shape}, density: {density_dict[name]}"
                )
                self.masks[name][:] = (torch.rand(mask.shape) < density_dict[name]).float().data.cuda()

                total_nonzero += density_dict[name] * mask.numel()
            print(f"Overall sparsity {total_nonzero / total_params}")

        elif mode == 'ERK':
            print('initialize by ERK')
            total_params = 0
            for name, weight in self.masks.items():
                total_params += weight.numel()
            is_epsilon_valid = False

            dense_layers = set()
            while not is_epsilon_valid:

                divisor = 0
                rhs = 0
                raw_probabilities = {}
                for name, mask in self.masks.items():
                    n_param = np.prod(mask.shape)
                    n_zeros = n_param * (1 - self.density)
                    n_ones = n_param * self.density

                    if name in dense_layers:
                        rhs -= n_zeros
                    else:
                        rhs += n_ones
                        raw_probabilities[name] = (
                                                          np.sum(mask.shape) / np.prod(mask.shape)
                                                  ) ** erk_power_scale
                        divisor += raw_probabilities[name] * n_param
                epsilon = rhs / divisor
                max_prob = np.max(list(raw_probabilities.values()))
                max_prob_one = max_prob * epsilon
                if max_prob_one > 1:
                    is_epsilon_valid = False
                    for mask_name, mask_raw_prob in raw_probabilities.items():
                        if mask_raw_prob == max_prob:
                            print(f"Sparsity of var:{mask_name} had to be set to 0.")
                            dense_layers.add(mask_name)
                else:
                    is_epsilon_valid = True

            density_dict = {}
            total_nonzero = 0.0
            # With the valid epsilon, we can set sparsities of the remaning layers.
            for name, mask in self.masks.items():
                n_param = np.prod(mask.shape)
                if name in dense_layers:
                    density_dict[name] = 1.0
                else:
                    probability_one = epsilon * raw_probabilities[name]
                    density_dict[name] = probability_one
                print(
                    f"layer: {name}, shape: {mask.shape}, density: {density_dict[name]}"
                )
                self.masks[name][:] = (torch.rand(mask.shape) < density_dict[name]).float().data.cuda()

                total_nonzero += density_dict[name] * mask.numel()
            print(f"Overall sparsity {total_nonzero / total_params}")

        self.apply_mask()

        if self.flow_preservation_flag:
            self.flow_preservation()

        total_size = 0
        for name, weight in self.masks.items():
            total_size += weight.numel()

        sparse_size = 0
        for name, weight in self.masks.items():
            sparse_size += (weight != 0).sum().int().item()
        print('Total parameters under sparsity level of {0}: {1}'.format(density, sparse_size / total_size))
        print('Sparsity at initialization: {0}'.format((total_size-sparse_size) / total_size))

        total_size = 0
        sparse_size = 0
        for module in self.modules:
            for name, weight in module.named_parameters():
                if name in self.masks:
                    print(name, 'density:',(weight!=0).sum().item()/weight.numel())
                    print(name, 'density:',(self.masks[name]!=0).sum().item()/self.masks[name].numel())
                    
                    total_size += weight.numel()
                    sparse_size += (weight != 0).sum().int().item()
        print('Total Model parameters:', total_size)
        print('Total parameters under sparsity level of {0}: {1}'.format(self.density, sparse_size / total_size))

        # self.fired_masks = copy.deepcopy(self.masks) # used for ITOP
        # self.print_nonzero_counts()


    def step(self):
        self.optimizer.step()
        self.apply_mask()
        self.death_rate_decay.step()
        self.death_rate = self.death_rate_decay.get_dr()
        self.steps += 1

        if self.prune_every_k_steps is not None:
            if self.steps % self.prune_every_k_steps == 0:
                self.truncate_weights()
                _, _ = self.fired_masks_update()
                self.print_nonzero_counts()


    def flow_preservation(self):
        layer = 0
        num_corrected = 0
        for module in self.modules:
            for name, weight in module.named_parameters():
                if name not in self.masks: continue
                mask = self.masks[name]
                
                if len(weight.shape) == 4:
                    s = mask.sum() / mask.numel()
                    active_filters = torch.sum(mask, (2,3)).clamp(max=1)
                    indegrees = active_filters.sum(1)
                    outdegrees = active_filters.sum(0)
                    idx_in = torch.where(indegrees == 0)[0]
                    idx_out = torch.where(outdegrees == 0)[0]
                    if (indegrees==0).any():
                        num_corrected += (indegrees==0).any().sum()
                        idx = torch.where(indegrees==0)[0]
                        temp = mask[idx, :, :, :]
                        for i in range(idx.shape[0]):
                            indices = torch.randperm(temp[i, :, :, :].numel())
                            temp[i, :, :, :].view(-1)[indices[0]] = 1
                            mask[idx[i], :, :, :] = temp[i, :, :, :]
                        active_filters = torch.sum(mask, (2,3)).clamp(max=1)
                        indegrees = active_filters.sum(1)
                        print('Corrected Flow Conv In', (indegrees==0).any(), mask.sum() / mask.numel(), s)
                    
                    if (outdegrees==0).any():
                        num_corrected += (outdegrees==0).any().sum()

                        idx = torch.where(outdegrees==0)[0]
                        temp = mask[:, idx, :, :]
                        for i in range(idx.shape[0]):
                            indices = torch.randperm(temp[:, i, :, :].numel())
                            temp[:, i, :, :].contiguous().view(-1)[indices[0]] = 1
                            mask[:, idx[i], :, :] = temp[:, i, :, :]
                        active_filters = torch.sum(mask, (2,3)).clamp(max=1)
                        outdegrees = active_filters.sum(1)
                        print('Corrected Flow Conv Out', (outdegrees==0).any(), mask.sum() / mask.numel(), s)
                        
                    layer += 1

                else:
                    s = mask.sum() / mask.numel()
                    outdegrees = mask.sum(0)
                    indegrees = mask.sum(1)
                    idx_in = torch.where(indegrees == 0)[0]
                    if (indegrees==0).any():
                        num_corrected += (indegrees==0).any().sum()

                        idx = torch.where(indegrees==0)[0]
                        temp = mask[idx, :]
                        for i in range(idx.shape[0]):
                            indices = torch.randperm(temp[i, :].numel())
                            temp[i, :].view(-1)[indices[0]] = 1
                            mask[idx[i], :] = temp[i, :]
                        indegrees = mask.sum(1)
                        print('Corrected Flow Linear in', (indegrees==0).any(), mask.sum() / mask.numel(), s)
                    if (outdegrees==0).any():
                        num_corrected += (outdegrees==0).any().sum()

                        idx = torch.where(outdegrees==0)[0]
                        temp = mask[:, idx]
                        for i in range(idx.shape[0]):
                            indices = torch.randperm(temp[:, i].numel())
                            temp[:, i].view(-1)[indices[0]] = 1
                            mask[:, idx[i]] = temp[:, i]
                        outdegrees = mask.sum(0)
                        print('Corrected Flow Linear out', (indegrees==0).any(), mask.sum() / mask.numel(), s)
                        
                    layer += 1
                self.masks[name] = mask
        self.num_corrected = num_corrected
        print("num_corrected = ", self.num_corrected)
        sv_path = './logs/num_corrected_{0}_{1}_{2}_{3}_{4}.txt'.format(self.args.model, self.args.density, self.args.sparse_init, self.args.seed, self.args.flow_preservation)
        file = open(sv_path, "w")
        str = repr(num_corrected)
        file.write("num_corrected = " + str + "\n")
        file.close()
        self.apply_mask()


    def add_module(self, module, density, sparse_init='ER'):
        self.modules.append(module)
        self.module = module
        for name, tensor in module.named_parameters():
            self.names.append(name)
            self.masks[name] = torch.zeros_like(tensor, dtype=torch.float32, requires_grad=False).cuda()

        print('Removing biases...')
        self.remove_weight_partial_name('bias')
        print('Removing 2D batch norms...')
        self.remove_type(nn.BatchNorm2d)
        print('Removing 1D batch norms...')
        self.remove_type(nn.BatchNorm1d)
        self.init(mode=sparse_init, density=density)


    def remove_weight(self, name):
        if name in self.masks:
            print('Removing {0} of size {1} = {2} parameters.'.format(name, self.masks[name].shape,
                                                                      self.masks[name].numel()))
            self.masks.pop(name)
        elif name + '.weight' in self.masks:
            print('Removing {0} of size {1} = {2} parameters.'.format(name, self.masks[name + '.weight'].shape,
                                                                      self.masks[name + '.weight'].numel()))
            self.masks.pop(name + '.weight')
        else:
            print('ERROR', name)

    def remove_weight_partial_name(self, partial_name):
        removed = set()
        for name in list(self.masks.keys()):
            if partial_name in name:

                print('Removing {0} of size {1} with {2} parameters...'.format(name, self.masks[name].shape,
                                                                                   np.prod(self.masks[name].shape)))
                removed.add(name)
                self.masks.pop(name)

        print('Removed {0} layers.'.format(len(removed)))

        i = 0
        while i < len(self.names):
            name = self.names[i]
            if name in removed:
                self.names.pop(i)
            else:
                i += 1

    def remove_type(self, nn_type):
        for module in self.modules:
            for name, module in module.named_modules():
                if isinstance(module, nn_type):
                    self.remove_weight(name)

    def apply_mask(self):
        for module in self.modules:
            for name, tensor in module.named_parameters():
                if name in self.masks:
                    # tensor = tensor.to(self.device)
                    tensor.data = tensor.data*self.masks[name]

    def truncate_weights(self):

        for module in self.modules:
            for name, weight in module.named_parameters():
                if name not in self.masks: continue
                mask = self.masks[name]
                self.name2nonzeros[name] = mask.sum().item()
                self.name2zeros[name] = mask.numel() - self.name2nonzeros[name]

                # death
                if self.death_mode == 'magnitude':
                    new_mask = self.magnitude_death(mask, weight, name)
                elif self.death_mode == 'SET':
                    new_mask = self.magnitude_and_negativity_death(mask, weight, name)
                elif self.death_mode == 'Taylor_FO':
                    new_mask = self.taylor_FO(mask, weight, name)
                elif self.death_mode == 'threshold':
                    new_mask = self.threshold_death(mask, weight, name)

                self.num_remove[name] = int(self.name2nonzeros[name] - new_mask.sum().item())
                self.masks[name][:] = new_mask


        for module in self.modules:
            for name, weight in module.named_parameters():
                if name not in self.masks: continue
                new_mask = self.masks[name].data.byte()

                # growth
                if self.growth_mode == 'random':
                    new_mask = self.random_growth(name, new_mask, weight)

                if self.growth_mode == 'random_unfired':
                    new_mask = self.random_unfired_growth(name, new_mask, weight)

                elif self.growth_mode == 'momentum':
                    new_mask = self.momentum_growth(name, new_mask, weight)

                elif self.growth_mode == 'gradient':
                    new_mask = self.gradient_growth(name, new_mask, weight)

                new_nonzero = new_mask.sum().item()

                # exchanging masks
                self.masks.pop(name)
                self.masks[name] = new_mask.float()

        self.apply_mask()


    '''
                    DEATH
    '''

    def threshold_death(self, mask, weight, name):
        return (torch.abs(weight.data) > self.threshold)

    def taylor_FO(self, mask, weight, name):

        num_remove = math.ceil(self.death_rate * self.name2nonzeros[name])
        num_zeros = self.name2zeros[name]
        k = math.ceil(num_zeros + num_remove)

        x, idx = torch.sort((weight.data * weight.grad).pow(2).flatten())
        mask.data.view(-1)[idx[:k]] = 0.0

        return mask

    def magnitude_death(self, mask, weight, name):

        num_remove = math.ceil(self.death_rate*self.name2nonzeros[name])
        if num_remove == 0.0: return weight.data != 0.0
        num_zeros = self.name2zeros[name]

        x, idx = torch.sort(torch.abs(weight.data.view(-1)))
        n = idx.shape[0]

        k = math.ceil(num_zeros + num_remove)
        threshold = x[k-1].item()

        return (torch.abs(weight.data) > threshold)


    def magnitude_and_negativity_death(self, mask, weight, name):
        num_remove = math.ceil(self.death_rate*self.name2nonzeros[name])
        num_zeros = self.name2zeros[name]

        # find magnitude threshold
        # remove all weights which absolute value is smaller than threshold
        x, idx = torch.sort(weight[weight > 0.0].data.view(-1))
        k = math.ceil(num_remove/2.0)
        if k >= x.shape[0]:
            k = x.shape[0]

        threshold_magnitude = x[k-1].item()

        # find negativity threshold
        # remove all weights which are smaller than threshold
        x, idx = torch.sort(weight[weight < 0.0].view(-1))
        k = math.ceil(num_remove/2.0)
        if k >= x.shape[0]:
            k = x.shape[0]
        threshold_negativity = x[k-1].item()


        pos_mask = (weight.data > threshold_magnitude) & (weight.data > 0.0)
        neg_mask = (weight.data < threshold_negativity) & (weight.data < 0.0)


        new_mask = pos_mask | neg_mask
        return new_mask

    '''
                    GROWTH
    '''

    def random_unfired_growth(self, name, new_mask, weight):
        total_regrowth = self.num_remove[name]
        n = (new_mask == 0).sum().item()
        if n == 0: return new_mask
        num_nonfired_weights = (self.fired_masks[name]==0).sum().item()

        if total_regrowth <= num_nonfired_weights:
            idx = (self.fired_masks[name].flatten() == 0).nonzero()
            indices = torch.randperm(len(idx))[:total_regrowth]

            # idx = torch.nonzero(self.fired_masks[name].flatten())
            new_mask.data.view(-1)[idx[indices]] = 1.0
        else:
            new_mask[self.fired_masks[name]==0] = 1.0
            n = (new_mask == 0).sum().item()
            expeced_growth_probability = ((total_regrowth-num_nonfired_weights) / n)
            new_weights = torch.rand(new_mask.shape).cuda() < expeced_growth_probability
            new_mask = new_mask.byte() | new_weights
        return new_mask

    def random_growth(self, name, new_mask, weight):
        total_regrowth = self.num_remove[name]
        n = (new_mask==0).sum().item()
        if n == 0: return new_mask
        expeced_growth_probability = (total_regrowth/n)
        new_weights = torch.rand(new_mask.shape).cuda() < expeced_growth_probability
        new_mask_ = new_mask.byte() | new_weights
        if (new_mask_!=0).sum().item() == 0:
            new_mask_ = new_mask
        return new_mask_

    def momentum_growth(self, name, new_mask, weight):
        total_regrowth = self.num_remove[name]
        grad = self.get_momentum_for_weight(weight)
        grad = grad*(new_mask==0).float()
        y, idx = torch.sort(torch.abs(grad).flatten(), descending=True)
        new_mask.data.view(-1)[idx[:total_regrowth]] = 1.0

        return new_mask

    def gradient_growth(self, name, new_mask, weight):
        total_regrowth = self.num_remove[name]
        grad = self.get_gradient_for_weights(weight)
        grad = grad*(new_mask==0).float()

        y, idx = torch.sort(torch.abs(grad).flatten(), descending=True)
        new_mask.data.view(-1)[idx[:total_regrowth]] = 1.0

        return new_mask



    def momentum_neuron_growth(self, name, new_mask, weight):
        total_regrowth = self.num_remove[name]
        grad = self.get_momentum_for_weight(weight)

        M = torch.abs(grad)
        if len(M.shape) == 2: sum_dim = [1]
        elif len(M.shape) == 4: sum_dim = [1, 2, 3]

        v = M.mean(sum_dim).data
        v /= v.sum()

        slots_per_neuron = (new_mask==0).sum(sum_dim)

        M = M*(new_mask==0).float()
        for i, fraction  in enumerate(v):
            neuron_regrowth = math.floor(fraction.item()*total_regrowth)
            available = slots_per_neuron[i].item()

            y, idx = torch.sort(M[i].flatten())
            if neuron_regrowth > available:
                neuron_regrowth = available
            threshold = y[-(neuron_regrowth)].item()
            if threshold == 0.0: continue
            if neuron_regrowth < 10: continue
            new_mask[i] = new_mask[i] | (M[i] > threshold)

        return new_mask

    '''
                UTILITY
    '''
    def get_momentum_for_weight(self, weight):
        if 'exp_avg' in self.optimizer.state[weight]:
            adam_m1 = self.optimizer.state[weight]['exp_avg']
            adam_m2 = self.optimizer.state[weight]['exp_avg_sq']
            grad = adam_m1/(torch.sqrt(adam_m2) + 1e-08)
        elif 'momentum_buffer' in self.optimizer.state[weight]:
            grad = self.optimizer.state[weight]['momentum_buffer']
        return grad

    def get_gradient_for_weights(self, weight):
        grad = weight.grad.clone()
        return grad

    def print_nonzero_counts(self):
        for module in self.modules:
            for name, tensor in module.named_parameters():
                if name not in self.masks: continue
                mask = self.masks[name]
                num_nonzeros = (mask != 0).sum().item()
                val = '{0}: {1}->{2}, density: {3:.3f}'.format(name, self.name2nonzeros[name], num_nonzeros, num_nonzeros/float(mask.numel()))
                print(val)


        for module in self.modules:
            for name, tensor in module.named_parameters():
                if name not in self.masks: continue
                print('Death rate: {0}\n'.format(self.death_rate))
                break

    def fired_masks_update(self):
        ntotal_fired_weights = 0.0
        ntotal_weights = 0.0
        layer_fired_weights = {}
        for module in self.modules:
            for name, weight in module.named_parameters():
                if name not in self.masks: continue
                self.fired_masks[name] = self.masks[name].data.byte() | self.fired_masks[name].data.byte()
                ntotal_fired_weights += float(self.fired_masks[name].sum().item())
                ntotal_weights += float(self.fired_masks[name].numel())
                layer_fired_weights[name] = float(self.fired_masks[name].sum().item())/float(self.fired_masks[name].numel())
                print('Layerwise percentage of the fired weights of', name, 'is:', layer_fired_weights[name])
        total_fired_weights = ntotal_fired_weights/ntotal_weights
        print('The percentage of the total fired weights is:', total_fired_weights)
        return layer_fired_weights, total_fired_weights
